r"""Contain plotting functions to analyze discrete values."""

from __future__ import annotations

__all__ = ["bar_discrete", "bar_discrete_temporal"]

from typing import TYPE_CHECKING

import numpy as np
from matplotlib import pyplot as plt

from flamme.plot.utils import auto_yscale_discrete, readable_xticklabels

if TYPE_CHECKING:
    from collections.abc import Sequence

    from matplotlib.axes import Axes


def bar_discrete(
    ax: Axes,
    names: Sequence,
    counts: Sequence[int],
    yscale: str = "auto",
) -> None:
    r"""Plot the histogram of an array containing discrete values.

    Args:
        ax: The axes of the matplotlib figure to update.
        names: The name of the values to plot.
        counts: The number of value occurrences.
        yscale: The y-axis scale. If ``'auto'``, the
            ``'linear'`` or ``'log'/'symlog'`` scale is chosen based
            on the distribution.

    Example usage:

    ```pycon

    >>> from matplotlib import pyplot as plt
    >>> from flamme.plot import bar_discrete
    >>> fig, ax = plt.subplots()
    >>> bar_discrete(ax, names=["a", "b", "c", "d"], counts=[5, 100, 42, 27])

    ```
    """
    n = len(names)
    if n == 0:
        return
    x = np.arange(n)
    ax.bar(x, counts, width=0.9 if n < 50 else 1, color="tab:blue")
    if yscale == "auto":
        yscale = auto_yscale_discrete(min_count=min(counts), max_count=max(counts))
    ax.set_yscale(yscale)
    ax.set_xticks(x, labels=map(str, names))
    readable_xticklabels(ax, max_num_xticks=100)
    ax.set_xlim(-0.5, len(names) - 0.5)
    ax.set_xlabel("values")
    ax.set_ylabel("number of occurrences")


def bar_discrete_temporal(
    ax: Axes,
    counts: np.ndarray,
    steps: Sequence | None = None,
    values: Sequence | None = None,
    proportion: bool = False,
) -> None:
    r"""Plot the temporal distribution of discrete values.

    Args:
        ax: The axes of the matplotlib figure to update.
        counts: A 2-d array that indicates the number of occurrences
            for each value and time step. The first dimension
            represents the value and the second dimension
            represents the steps.
        steps: The name associated to each step.
        values: The name associated to each value.
        proportion: If ``True``, it plots the normalized number of
            occurrences for each step.

    Example usage:

    ```pycon

    >>> from matplotlib import pyplot as plt
    >>> from flamme.plot import bar_discrete_temporal
    >>> fig, ax = plt.subplots()
    >>> bar_discrete_temporal(
    ...     ax, counts=np.ones((5, 20)), values=list(range(5)), steps=list(range(20))
    ... )

    ```
    """
    if counts.size == 0:
        return
    num_values, num_steps = counts.shape
    values = _prepare_values_bar_discrete_temporal(values=values, num_values=num_values)
    steps = _prepare_steps_bar_discrete_temporal(steps=steps, num_steps=num_steps)
    counts = _prepare_counts_bar_discrete_temporal(counts=counts, proportion=proportion)

    x = np.arange(num_steps, dtype=np.int64)
    bottom = np.zeros(num_steps, dtype=counts.dtype)
    width = 0.9 if num_steps < 50 else 1
    my_cmap = plt.get_cmap("viridis")
    for i in range(num_values):
        count = counts[i]
        ax.bar(x, count, label=values[i], bottom=bottom, width=width, color=my_cmap(i / num_values))
        bottom += count

    num_valid_values = len(list(filter(lambda x: x is not None, values)))
    if num_valid_values <= 10 and num_valid_values > 0:
        ax.legend()
    ax.set_xticks(x, labels=steps)
    readable_xticklabels(ax, max_num_xticks=100)
    ax.set_xlim(-0.5, num_steps - 0.5)
    ax.set_ylabel("steps")
    ax.set_ylabel("proportion" if proportion else "number of occurrences")


def _prepare_values_bar_discrete_temporal(values: Sequence | None, num_values: int) -> list:
    r"""Return the list of values.

    This function was designed to be used in ``bar_discrete_temporal``.

    Args:
        values: The sequence of values.
        num_values: The expected number of values.

    Returns:
        The values. If ``values`` is ``None``, a list filled with
            ``None`` is returned.

    Raises:
        RuntimeError: if the length of ``values`` does not match with
            ``num_values``.
    """
    if values is None:
        return [None] * num_values
    if len(values) != num_values:
        msg = (
            f"values length ({len(values):,}) do not match with the count matrix "
            f"first dimension ({num_values:,})"
        )
        raise RuntimeError(msg)
    return list(values)


def _prepare_steps_bar_discrete_temporal(steps: Sequence | None, num_steps: int) -> list:
    r"""Return the list of steps.

    This function was designed to be used in ``bar_discrete_temporal``.

    Args:
        steps: The sequence of steps.
        num_steps: The expected number of steps.

    Returns:
        The steps. If ``steps`` is ``None``, a list filled with
            ``None`` is returned.

    Raises:
        RuntimeError: if the length of ``steps`` does not match with
            ``num_steps``.
    """
    if steps is None:
        return list(range(num_steps))
    if len(steps) != num_steps:
        msg = (
            f"steps length ({len(steps):,}) do not match with the count matrix "
            f"second dimension ({num_steps:,})"
        )
        raise RuntimeError(msg)
    return list(steps)


def _prepare_counts_bar_discrete_temporal(counts: np.ndarray, proportion: bool) -> np.ndarray:
    r"""Prepare the count matrix.

    This function was designed to be used in ``bar_discrete_temporal``.

    Args:
        counts: A 2-d array that indicates the number of occurrences
            for each value and time step. The first dimension
            represents the value and the second dimension
            represents the steps.
        proportion: If ``True``, the count matrix is normalized number
            of occurrences for each step.

    Returns:
        The count matrix.
    """
    if not proportion:
        return counts
    return counts / np.clip(counts.sum(axis=0), a_min=1, a_max=None)



r"""Implement an analyzer to analyze only a subset of the columns."""

from __future__ import annotations

__all__ = ["ColumnSubsetAnalyzer"]

import logging
from typing import TYPE_CHECKING

from coola.utils import str_indent, str_mapping

from flamme.analyzer.base import BaseAnalyzer
from flamme.utils import setup_object

if TYPE_CHECKING:
    from collections.abc import Sequence

    import polars as pl

    from flamme.section import BaseSection

logger = logging.getLogger(__name__)


class ColumnSubsetAnalyzer(BaseAnalyzer):
    r"""Implement an analyzer to analyze only a subset of the columns.

    Args:
        columns: Soecifies the columns to select.
        analyzer: The analyzer
            or its configuration.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> import polars as pl
    >>> from flamme.analyzer import ColumnSubsetAnalyzer, NullValueAnalyzer
    >>> analyzer = ColumnSubsetAnalyzer(columns=["float", "str"], analyzer=NullValueAnalyzer())
    >>> analyzer
    ColumnSubsetAnalyzer(
      (columns): 2 ['float', 'str']
      (analyzer): NullValueAnalyzer(figsize=None)
    )
    >>> frame = pl.DataFrame(
    ...     {
    ...         "float": [1.2, 4.2, None, 2.2],
    ...         "int": [None, 1, 0, 1],
    ...         "str": ["A", "B", None, None],
    ...     },
    ...     schema={"float": pl.Float64, "int": pl.Int64, "str": pl.String},
    ... )
    >>> section = analyzer.analyze(frame)
    >>> section
    NullValueSection(
      (columns): ('float', 'str')
      (null_count): array([1, 2])
      (total_count): array([4, 4])
      (figsize): None
    )

    ```
    """

    def __init__(self, columns: Sequence[str], analyzer: BaseAnalyzer | dict) -> None:
        self._columns = list(columns)
        self._analyzer = setup_object(analyzer)

    def __repr__(self) -> str:
        args = str_indent(
            str_mapping(
                {"columns": f"{len(self._columns)} {self._columns}", "analyzer": self._analyzer}
            )
        )
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def analyze(self, frame: pl.DataFrame) -> BaseSection:
        logger.info(f"Selecting {len(self._columns):,} columns: {self._columns}")
        return self._analyzer.analyze(frame.select(self._columns))



r"""Contain utility functions to generate data in a temporal range."""

from __future__ import annotations

__all__ = ["datetime_range"]


from typing import TYPE_CHECKING, Any

import polars as pl
from grizz.utils.datetime import find_end_datetime

if TYPE_CHECKING:
    from datetime import date, datetime, timedelta

    from polars._typing import IntoExprColumn


def datetime_range(
    start: datetime | date | IntoExprColumn,
    periods: int,
    interval: str | timedelta = "1d",
    **kwargs: Any,
) -> pl.Series | pl.Expr:
    r"""Generate a datetime range.

    Args:
        start: The lower bound of the datetime range.
        interval: The interval of the range periods, specified as a
            Python timedelta object or using the Polars duration
            string language.
        periods: The number of periods after the start.
        **kwargs: Keyword arguments passed to ``polars.datetime_range``

    Returns:
        Column of data type ``Datetime``.

    Example usage:

    ```pycon

    >>> from datetime import datetime, timezone
    >>> from flamme.utils.data import datetime_range
    >>> series = datetime_range(
    ...     start=datetime(
    ...         year=2017, month=2, day=3, hour=4, minute=5, second=6, tzinfo=timezone.utc
    ...     ),
    ...     periods=5,
    ...     interval="1d",
    ...     eager=True,
    ... ).alias("datetime")
    >>> series
    shape: (5,)
    Series: 'datetime' [datetime[μs, UTC]]
    [
        2017-02-03 04:05:06 UTC
        2017-02-04 04:05:06 UTC
        2017-02-05 04:05:06 UTC
        2017-02-06 04:05:06 UTC
        2017-02-07 04:05:06 UTC
    ]

    ```
    """
    return pl.datetime_range(
        start=start,
        end=find_end_datetime(
            start=start,
            periods=periods - 1,
            interval=interval,
        ),
        interval=interval,
        **kwargs,
    )
